"""Manipulability objective function and its gradient

Functions
---------
manipulability(joint_position)
manipulability_gradient(joint_position)

"""

import numpy as np
from numpy.linalg import det
from scipy.optimize import approx_fprime
from kinova_gen3.kinematics.jacobian import jacobian


def manipulability(joint_position):
    """Manipulability objective function

    Calculate the cost of the manipulatility for a given joint configuration

    Arguments
    ---------
    joint_position (array_like): Joint positions of the robot [rad]

    Returns
    -------
    float: Manipulability of the robot [non-dimensional]

    """

    return np.log(det(jacobian(joint_position) @ jacobian(joint_position).transpose()))


def manipulability_gradient(joint_position):
    """Gradient of the manipulability objective function

    Arguments
    ---------
    joint_position (array_like): Joint positions of the robot [rad]

    Returns
    -------
    ndarray: Gradient of the manipulability of the robot [non-dimensional]

    """

    return approx_fprime(joint_position, manipulability, np.sqrt(np.finfo(float).eps))



'''Test the forward kinematics of Kinova Gen3

Classes
-------
TestForwardKinematics

Functions
---------
test_forward_kinematics()

'''

import math
import numpy as np
import numpy.testing as npt
import unittest
from kinova_gen3.forward_kinematics import forward_kinematics


class TestForwardKinematics(unittest.TestCase):
    '''Unit test class for testing the forward kinematics of Kinova Gen3

    Methods
    -------
    test_forward_kinematics()
        Test the position level forward kinematics

    '''

    def test_forward_kinematics(self):
        '''Test the position level forward kinematics of Kinova Gen3'''

        # Test case 1: all joint positions are zero
        joint_pos = np.zeros(7)

        npt.assert_array_equal(forward_kinematics(joint_pos)[0], [0. , -0.0246, 1.1873])
        npt.assert_array_equal(forward_kinematics(joint_pos)[1], np.eye(3))

        # Test case 2: Rotate second joint by pi/2
        joint_pos[1] = math.pi / 2

        npt.assert_allclose(forward_kinematics(joint_pos)[0], [0.9025 , -0.0246, 0.2848])
        npt.assert_allclose(forward_kinematics(joint_pos)[1], \
                np.array([[0.0, 0.0, 1.0], [0.0, 1.0, 0.0], [-1.0, 0.0, 0.0]]), 1e-6, 1e-7)



"""Joint limit objective function and its gradient

Functions
---------
joint_limits(joint_position)
joint_limits_gradient(joint_position)

"""

import numpy as np
from scipy.optimize import approx_fprime


def joint_limits(joint_position):
    """Joint limits objective function

    Calculate the cost of the current joint configuration from the lower
    and upper bounds of the joints

    Arguments
    ---------
    joint_position (array_like): Joint position of the robot [rad]

    Returns
    -------
    float: Cost of the current configuration [non-dimensional]

    """

    joint_position_min = (
        np.array([-180.0, -128.9, -180.0, -147.8, -180.0, -120.3, -180.0])
        * np.pi
        / 180.0
    )
    joint_position_max = (
        np.array([180.0, 128.9, 180.0, 147.8, 180.0, 120.3, 180.0]) * np.pi / 180.0
    )

    joint_position_bar = (joint_position_min + joint_position_max) / 2

    return (-1 / (2 * joint_position.shape[0])) * np.sum(
        (
            (joint_position - joint_position_bar)
            / (joint_position_max - joint_position_min)
        )
        ** 2
    )


def joint_limits_gradient(joint_position):
    """Gradient of the joint limits objective function

    Calculate the first derivative of the joint limits objective
    function with respect to all joint variables

    Arguments
    ---------
    joint_position (array_like): Joint position of the robot [rad]

    Returns
    -------
    ndarray: Gradient vector for the joint limits objective function [non-dimensional]

    """

    return approx_fprime(joint_position, joint_limits, np.sqrt(np.finfo(float).eps))
